"""
Full-featured e-mailing system: flexible, slim and sexy.

* Unicode
* Easy attachments
* Inline images
* E-Mail templates
* Tools for unit-tests
* Made perfect once and for all. Simple and cute :)

Here'a a full example:

```python
from mailem import Message, Postman, Attachment, ImageAttachment
from mailem.connection import SMTPConnection

# Create the message
messages = [
    # Message with attachments
    Message(
        ['kolypto@gmail.com'],
        u"Mail'em test",
        u"<b>yeah baby, it works!</b>",
        attachments = [
            Attachment(u'test.txt', open('test.txt').read())
        ]
    ),
    # Message with inline images (!)
    Message(
        ['kolypto@gmail.com'],
        u"Mail'em test with inline images",
        u"Cute: <img src='cid:cute.jpg' />",  # cid:<filename>
        attachments = [
            ImageAttachment('cute.jpg', open('cute.jpg').read(), 'inline')
        ]
    ),
]

# Initialize a postman with SMTP connection to GMail
postman = Postman('user@gmail.com',
                  SMTPConnection(
                      'smtp.gmail.com', 587,
                      'user@gmail.com', 'pass',
                      tls=True
                  ))

# Send everything we have
with postman.connect() as c:
    map(c.sendmail, messages)
```

Also see [Template](#template).
"""

from .message import Message
from .attachment import Attachment, ImageAttachment
from .postman import Postman
from . import connection, template



import itertools

from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formatdate, make_msgid

from .util import Address, unicode_header


class Message(object):
    """ Construct a Message object.

    Notes:

    * Full unicode support, and Unicode is the default
    * You can provide `html` or `text` contents. If both are specified -- the message will have an 'alternative' container,
      so the user will receive both HTML and plaintext. The client will choose which one to display.
    * E-Mail addresses, such as `recipients` and `sender`, can be specified in one of the following formats:

        * `'user@example.com'`: Just an e-mail address
        * `('user@example.com', u'Honored User')`: email address with name

    Arguments:

    :param recipients: List of recipients
    :type recipients: Iterable[basestring|tuple[basestring]]
    :param subject: Message subject
    :type subject: basestring
    :param html: Message body, HTML
    :type html: basestring|None
    :param text: Message body, Text
    :type text: basestring|None
    :param sender: Sender e-mail address. If not set explicitly, the default will be used on send
    :type sender: basestring|tuple[basestring]|None
    :param cc: CC list
    :type cc: Iterable[basestring|tuple[basestring]]|None
    :param bcc: BCC list
    :type bcc: Iterable[basestring|tuple[basestring]]|None
    :param attachments: List of attachments
    :type attachments: Iterable[mailem.attachment.Attachment]
    :param reply_to: Reply-to address
    :type reply_to: basestring|tuple[basestring]|None
    :param date: Send date
    :type date: datetime|None
    :param headers: Additional headers
    :type headers: dict
    """

    def __init__(self, recipients, subject, html=None, text=None, sender=None, cc=None, bcc=None, attachments=None, reply_to=None, date=None, headers=None):
        self._recipients = [Address(r) for r in recipients]
        self._subject = subject
        self._html = html
        self._text = text
        self._sender = Address(sender) if sender else None
        self._cc =  [Address(c) for c in cc or ()]
        self._bcc = [Address(b) for b in bcc or ()]
        self._attachments = attachments or []
        self._reply_to = Address(reply_to) if reply_to else None
        self._date = date
        self._headers = headers or {}
        self._msgid = make_msgid()

    def _sender_default(self, sender):
        """ Set the default sender address

        :param sender: Sender address
        :type sender: basestring|tuple[basestring]
        """
        if not self._sender:
            self._sender = Address(sender)

    def _mime(self):
        """ Build a MIME object for this message

        :return:
        :rtype: email.mime.text.MIMEText|email.mime.multipart.MIMEMultipart
        """
        # Text object
        text = [ MIMEText(data, type,  'utf-8')
                 for data, type in (
                     (self._html, 'html'),
                     (self._text, 'plain'))
                 if data is not None]
        if len(text) == 1:
            msg = text[0]
        else:
            msg = MIMEMultipart('alternative')
            map(msg.attach, text)

        # Attachments
        if self._attachments:
            _text_msg = msg
            msg = MIMEMultipart()
            msg.attach(_text_msg)
            map(msg.attach, (a._mime() for a in self._attachments))

        # Fields
        msg['Subject'] = unicode_header(self._subject)  # special

        # Headers
        headers = dict(self._headers)
        headers['Date'] = formatdate(self._date)  # handles `None` correctly
        headers['Message-ID'] = self._msgid
        headers.update({  # Address lists
            key: ', '.join(map(str, addresses))
            for key, addresses in (
                ('To', self._recipients),
                ('Cc', self._cc),
                ('Bcc', self._bcc),
                # Not lists, but added here (DRY)
                ('From', (self._sender,) if self._sender else None),
                ('Reply-To', (self._reply_to,) if self._reply_to else None))
            if addresses
        })

        # Finish
        for k, v in headers.items():
            msg[k] = v
        #itertools.starmap(msg.add_header, headers.items())
        return msg

    def __str__(self):
        """ Build the MIME object and get a string """
        return self._mime().as_string()



import mailem
from exdoc import doc, getmembers

import json

doccls = lambda cls, *predicates: {
    'cls': doc(cls),
    'attrs': {name: doc(m) for name, m in getmembers(cls, *predicates)}
}

data = {
    'mailem': doc(mailem),
    'Message': doc(mailem.Message),
    'Attachment': doc(mailem.Attachment),
    'ImageAttachment': doc(mailem.ImageAttachment),
    'Postman': doccls(mailem.Postman),
    'connection': doc(mailem.connection),
    'SMTPConnection': doc(mailem.connection.SMTPConnection),
    'LoopbackConnection': doc(mailem.connection.LoopbackConnection),
    'Template': doccls(mailem.template.Template, None, lambda k, v: k=='__call__' or not k.startswith('_')),
}

print json.dumps(data, indent=2)



import smtplib
import itertools

from .base import IConnection


class SMTPConnection(IConnection):
    """ SMTP connection.

    See [smtplib](https://docs.python.org/2/library/smtplib.html) for the list of exceptions that may occur.

    Example:

    ```python
    from mailem import Postman
    from mailem.connection import SMTPConnection

    postman = Postman('user@gmail.com',
                  SMTPConnection(
                      'smtp.gmail.com', 587,
                      'user@gmail.com', 'pass',
                      tls=True
                  ))

    with postman.connect() as c:
        c.sendmail(msg)
    ```

    Arguments:

    :param host: SMTP server hostname
    :type host: str
    :param port: SMTP server port number.
    :type port: int
    :param username: User name to authenticate with
    :type username: str
    :param password: Password
    :type password: str
    :param local_hostname: FQDN of the local host for the HELO/EHLO command. When `None`, is detected automatically.
    :type local_hostname: str|None
    :param ssl: Use SSL protocol?
    :type ssl: bool
    :param tls: Use TLS handshake?
    :type tls: bool
    """

    def __init__(self, host, port, username, password, local_hostname=None, ssl=False, tls=False):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.local_hostname = local_hostname
        self.ssl = ssl
        self.tls = tls

        self.client = None

    def connect(self):
        # Init
        s = (smtplib.SMTP_SSL
             if self.ssl else
             smtplib.SMTP)(self.host, self.port, self.local_hostname)

        # Handshake
        if self.tls:
            s.starttls()
        s.login(self.username, self.password)

        # Finish
        self.client = s

    def disconnect(self):
        self.client.quit()
        self.client = None

    def sendmail(self, message):
        self.client.sendmail(
            # From
            message._sender.email,

            # To
            [r.email for r in itertools.chain(
                message._recipients,
                message._cc,
                message._bcc)],

            # Message
            str(message)
        )



""" Helpers & Utils """

from email import charset
from email.header import Header
from email.utils import formataddr


# Add charset to the global registry, since we're only using unicode here
# NOTE: global effect!
charset.add_charset('utf-8', charset.SHORTEST, None, 'utf-8')


def unicode_header(*args, **kwargs):
    """ Shortcut to create a unicode Header() string.

    :rtype: str
    """
    return str(Header(*args, charset='utf-8', **kwargs))


class Address(object):
    """ E-Mail address

    :param arg: e-mail address, or a 2-tuple (email, name)
    :type arg: basestring|tuple[basestring]
    """

    def __init__(self, arg):
        if isinstance(arg, basestring):
            self.email = arg
            self.name = None
        elif isinstance(arg, tuple) and len(arg) == 2:
            self.email, self.name = arg
        else:
            raise ValueError('Invalid address: should be a string or a 2-tuple, having: {!r}'.format(arg))

    def __hash__(self):
        return hash((self.email, self.name))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return (self.email, self.name) == (other.email, other.name)

    def __str__(self):
        return formataddr((
            unicode_header(self.name) if self.name else None,
            self.email
        ))

    def __repr__(self):
        if not self.name:
            return u'Address({!r})'.format(self.email)
        return u'Address(({!r}, {!r}))'.format(self.email, self.name)



import os
import re

from .. import Message, Attachment
from .renderer import PythonTemplateRenderer


class Template(object):
    """ A templated e-mail.

    By default, the Template uses Python's `Template` renderer, which allows simple PHP-style substitution,
    but this can be overridden using set_renderer().

    First, a template is defined:

    ```python
    from mailem import Attachment
    from mailem.template import Template

    signup = Template('Congrats $user, you\'ve signed up!',
        'Welcome to our website!<br><img src="cid:logo.jpg" /> -- $domain',
        attachments=[
            Attachment('logo.jpg', open('logo.jpg').read(), 'inline'))
        ],
        defaults={'domain': 'localhost'}  # default template values
    )
    ```

    Now, having the template, you render it to a [`Message`](#message) by calling it:

    ```python
    message = signup(['user@gmail.com'], dict(user='Honored User',))
    ```

    Ready for sending! :)

    :param subject: Message subject template
    :type subject: basestring|None
    :param html: HTML message template, if any
    :type html: basestring|None
    :param text: Text message template, if any
    :type text: basestring|None
    :param attachments: Attachments for the template. Most probably, inline elements.
    :type attachments: Iterable[mailem.Attachment]|None
    :param defaults: Default template values, if required. The user can override these later.
    :type defaults: dict|None
    """

    @classmethod
    def from_directory(cls, path, subject_name='subject.txt', html_name='index.htm', text_name='index.txt', inline_rex=r'^i-(.*)'):
        """ Convenience class method to import a directory as a template:

        * `subject.txt` is the subject string template
        * `index.htm` is the HTML template
        * `index.txt` is the plaintext template
        * All files matching the 'i-(*)' format are attached as 'inline', and hence can be referenced in the template:

            E.g. file 'i-flower.jpg' can be inlined as `<img src="cid:flower.jpg" />`.

        * All other files are just attachments.

        Example:

        ```python
        signup = Template.from_directory('templates/signup/')
        ```

        :param path: Path to the directory
        :type path: str
        :param subject_name: Subject template filename
        :type subject_name: str
        :param html_name: Html template filename
        :type html_name: str
        :param text_name: Plaintext template filename
        :type text_name: str
        :param inline_rex: Regular expression to match files that should be inlined.

            If the RegExp defines capture groups, group $1 will be used as the fact filename.

        :returns: Template
        :rtype: Template
        """
        subject = None
        html = None
        text = None
        attachments = []

        # Find files
        for filename in os.listdir(path):
            # Ignore hidden files
            if filename.startswith('.'):
                continue
            fpath = os.path.join(path, filename)

            # Ignore directories
            if os.path.isdir(fpath):
                continue

            # Get content
            with open(fpath) as f:
                content = f.read()

            # Place
            if filename == subject_name:
                subject = content.strip()  # No \r\n allowed here
            elif filename == html_name:
                html = content
            elif filename == text_name:
                text = content
            else:
                # Match filename
                m = re.match(inline_rex, filename)
                if m is None:
                    # Attachment
                    disposition = 'attachment'
                else:
                    # Inline attachment
                    disposition = 'inline'
                    try:
                        # Has capture groups
                        filename = m.group(1)
                    except IndexError:
                        # Has no capture groups
                        filename = m.group(0)

                # Attach
                attachments.append(Attachment(filename, content, disposition=disposition))

        # Template
        return cls(subject=subject, html=html, text=text, attachments=attachments)

    def __init__(self, subject=None, html=None, text=None, attachments=None, defaults=None):
        self._subject = subject
        self._html = html
        self._text = text
        self._attachments = attachments
        self._default_values = {}
        self._renderer = None
        self.defaults(defaults or {})

    def defaults(self, values):
        """ Set default values

        :param values: Default template values
        :type values: dict
        """
        self._default_values.update(values)
        return self

    def set_renderer(self, Renderer):
        """ Set renderer to be used with this template.

        A Renderer is any class that can be constructed with a template string argument,
        and called with template values dict to render it.

        When no renderer was explicitly set, it defaults to PythonTemplateRenderer.

        :param Renderer: Renderer class.
        :type Renderer: type
        """
        assert self._renderer is None, 'Cannot re-assign a renderer'

        self._renderer = Renderer
        self._subject = Renderer(self._subject)
        self._html = Renderer(self._html) if self._html else None
        self._text = Renderer(self._text) if self._text else None

        return self

    def __call__(self, recipients, values, **kwargs):
        """ Create a `Message` object using the template values.

        :param recipients: Message recipients list
        :type recipients: Iterable[basestring|tuple[basestring]]
        :param values: Dictionary with template values
        :type values: dict
        :param kwargs: keyword arguments for the [`Message`](#message) constructor
        :type kwargs: dict
        :returns: The rendered `Message` object
        :rtype: Message
        :raises KeyError: A template value was not provided
        :raises Exception: other renderer exceptions
        """
        # Default renderer
        if self._renderer is None:
            self.set_renderer(PythonTemplateRenderer)

        values = dict(self._default_values.items() + values.items())

        return Message(
            recipients,
            subject=self._subject(values),
            html=self._html(values) if self._html else None,
            text=self._text(values) if self._text else None,
            attachments=list(self._attachments or []) + list(kwargs.pop('attachments', [])),
            **kwargs
        )



from .connection.lo import LoopbackConnection


class Postman(object):
    """ Postman is the object you use to send messages through a configured Connection object.

    Example:

    ```python
    from mailem import Message, Postman
    from mailem.connection import SMTPConnection

    # Construct the message
    msg = Message(
        ['kolypto@gmail.com'],
        u"Mail'em test",
        u"<b>yeah baby, it works!</b>"
    )

    # Create the postman (see SMTPConnection)
    postman = Postman('user@gmail.com',
                      SMTPConnection(...))

    # Connect, and send the message
    with postman.connect() as c:
        c.sendmail(msg)
    ```

    :param sender: Default sender: e-mail or (name, email).
        Is used for messages which do not specify the sender address explicitly.
    :type sender: basestring|tuple[basestring]
    :param connection: Connection object to use. See below.
    :type connection: mailem.connection.IConnection
    """

    def __init__(self, sender, connection):
        self._sender = sender
        self._connection = connection

    def connect(self):
        """ Get connected Postman context manager.

        :rtype: mailem.postman.ConnectedPostman
        """
        return ConnectedPostman(self._sender, self._connection)

    def loopback(self):
        """ Get a context manager which installs a LoopbackConnection on this postman.

        This allows you to record outgoing messages by mocking a Postman.
        See [`LoopbackConnection`](#loopbackconnection).

        :return: Context manager which loops back outgoing messages
        :rtype: MockedPostman
        """
        return MockedPostman(self)


class ConnectedPostman(Postman):
    def __init__(self, *args):
        super(ConnectedPostman, self).__init__(*args)

    def sendmail(self, message):
        """ Send the message

        :param message: Message
        :type message: mailem.message.Message
        :return: The same message
        :rtype: mailem.message.Message
        """
        message._sender_default(self._sender)
        self._connection.sendmail(message)
        return message

    def __enter__(self):
        """ Connect """
        self._connection.connect()
        self._connected = True
        return self

    def __exit__(self, *exc):
        """ Disconnect """
        self._connection.disconnect()
        self._connected = False


class MockedPostman(LoopbackConnection):
    """ Mocks Postman with a loopback

    :param postman: Postman to mock
    :type postman: mailem.postman.Postman
    """

    def __init__(self, postman):
        self._postman = postman
        self._connection_back = None
        super(MockedPostman, self).__init__()

    def __enter__(self):
        # Backup
        self._connection_back = self._postman._connection
        # Mock with self
        self._postman._connection = self
        return self

    def __exit__(self, *exc):
        # Restore backup
        self._postman._connection = self._connection_back



""" Attachments """

import itertools
from urllib import quote_plus

from email.encoders import encode_base64
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage

from .util import unicode_header


class Attachment(object):
    """ File attachment information.

    This can be provided to the [`Message`](#message) object on construction.

    :param filename: Filename of attachment
    :type filename: str|unicode|None
    :param data: Taw file data
    :type data: str|None
    :param content_type: File mimetype
    :type content_type: str|None
    :param disposition: Content-Disposition: 'attachment', 'inline', ...
    :type disposition: str|None
    :param headers: Additional headers for the attachment
    :type headers: dict|None
    """

    def __init__(self, filename, data, content_type='application/octet-stream', disposition='attachment', headers=None):
        self.filename = filename
        self.content_type = content_type
        self.data = data
        self.disposition = disposition
        self.headers = headers or {}

        # Inlining
        if self.disposition == 'inline':
            self.headers.setdefault('Content-ID', '<{}>'.format(quote_plus(filename)))

    def _build_mime_object(self):
        """ Create a MIMe object

        :rtype: email.mime.base.MIMEBase
        """
        obj = MIMEBase(*self.content_type.split('/'))
        obj.set_payload(self.data)
        encode_base64(obj)
        return obj

    def _mime(self):
        """ Build a MIME object for the attachment

        :return: MIMEBase
        :rtype: email.mime.base.MIMEBase
        """
        obj = self._build_mime_object()

        # Content-Disposition
        obj.add_header('Content-Disposition', self.disposition, filename=unicode_header(self.filename))

        # Add headers
        for k, v in self.headers.items():
            obj[k] = v
        #itertools.starmap(obj.add_header, self.headers.items())  # does not work

        # Finish
        return obj


class ImageAttachment(Attachment):
    """ Image attachment.

    * It guesses the Content-Type from the data stream
    * Supports 'inline' images: images embedded in the email. Useful for templates.

        Once an 'inline' image is created, its filename is used for 'Content-ID', which allows to reference it in the HTML body:

        ```python
        from mailem import Message, Attachment, ImageAttachment

        msg = Message(
            ['test@example.com'],
            'Hello',
            '<img src="cid:flowers.jpg" />',  # Referenced with "cid:<filename>"
            attachments=[
                ImageAttachment('flowers.jpg', open('flowers.jpg').read(), 'inline')
            ]
        )
        ```

    Arguments:

    :param filename: Image attachment filename. Will also become 'Content-ID' when inlined.
    :type filename: str|unicode|None
    :param data: The raw file data
    :type data: str|None
    """

    def __init__(self, filename, data, disposition='attachment', headers=None):
        super(ImageAttachment, self).__init__(filename, data, None, disposition, headers)

    def _build_mime_object(self):
        return MIMEImage(self.data)



from .base import IConnection


class LoopbackConnection(list, IConnection):
    """ Loopback connection allows to record all outgoing messages instead of sending them.

    You can install it manually:

    ```python
    from mailem import Postman
    from mailem.connection import LoopbackConnection

    lo = LoopbackConnection()
    postman = Postman('user@example.com', lo)
    #... send
    messages = lo.get_messages()
    ```

    or you can mock an existing Postman with `loopback()` helper:

    ```python
    from mailem import Postman
    from mailem.connection import SMTPConnection

    postman = Postman('user@example.com',
                  SMTPConnection(...))

    with postman.loopback() as lo:
        # Send
        with postman.connect() as c:  # mocked!
            c.sendmail(msg)

    # Get
    sent_messages = lo.get_messages()
    ```

    Loopback can be installed multiple times, and only top-level loopback will catch the messages:

    ```python
    with postman.loopback() as lo1:
        with postman.loopback() as lo2:
            with postman.connect() as c:
                c.sendmail(msg)

    len(lo1)  #-> 0
    len(lo2)  #-> 1
    ```

    Also note that `LoopbackConnection` subclasses `list`, so all list methods, including iteration, is available.
    """

    def __init__(self):
        super(LoopbackConnection, self).__init__()

    def connect(self):
        pass

    def disconnect(self):
        pass

    def sendmail(self, message):
        self.append(message)

    #region Access messages

    def get_messages(self):
        """ Get all messages and reset the internal queue

        :return: List of recorded messages
        :rtype: list[mailem.Message]
        """
        return [self.pop() for i in range(len(self))]

    #endregion



#!/usr/bin/env python
""" Full-featured e-mailing system: flexible, slim and sexy """

from setuptools import setup, find_packages

setup(
    # http://pythonhosted.org/setuptools/setuptools.html
    name='mailem',
    version='0.0.1-0',
    author='Mark Vartanyan',
    author_email='kolypto@gmail.com',

    url='https://github.com/kolypto/py-mailem',
    license='BSD',
    description=__doc__,
    long_description=open('README.rst').read(),
    keywords=['e-mail', 'mail', 'template'],

    packages=find_packages(),
    scripts=[],
    entry_points={},

    install_requires=[
    ],
    extras_require={
        '_dev': ['wheel', 'nose', 'exdoc'],
    },
    include_package_data=True,
    test_suite='nose.collector',

    platforms='any',
    classifiers=[
        # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        #'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



class IConnection(object):
    """ Interface for connection objects """

    def connect(self):
        """ Establish a connection """
        raise NotImplementedError()

    def disconnect(self):
        """ Terminate the connection """
        raise NotImplementedError()

    def sendmail(self, message):
        """ Send the message

        :type msg: Message
        """
        raise NotImplementedError()



from string import Template


class IRenderer(object):
    """ Base class for template renderer """

    def __init__(self, template):
        """ Init renderer with a template string

        :param template: Template string
        :type template: basestring
        """
        raise NotImplementedError()

    def __call__(self, values):
        """ Render a template with the provided values.

        :param values: Values to render the template with.
        :type values: dict
        :returns: Rendered template
        :rtype: basestring
        """
        raise NotImplementedError()


class PythonTemplateRenderer(IRenderer):
    """ Simple Python Template renderer.

        Supported substitutions:

        * PythonTemplateRenderer('$what')(what=1)  #-> '1'
        * PythonTemplateRenderer('${what}')(what=1)  #-> '1'
        * PythonTemplateRenderer('$$what')(what=1)  #-> '$what'
    """

    def __init__(self, template):
        self.template = Template(template)

    def __call__(self, values):
        return self.template.substitute(values)
